"""
CryptoPix V4 - Ultra-Fast Core Encryption Module

This module implements the revolutionary CryptoPix V4 specification with:
- Ultra-fast encryption (<5ms performance target)
- Vectorized operations using NumPy with SIMD instructions
- Pre-computed lookup tables for instant color transformations
- Optimized PBKDF2 with adaptive iterations
- Hardware-accelerated AES when available
- Streamlined metadata with minimal overhead
- Post-quantum resistance through symmetric cryptography
- Backward compatibility with V2 keys
"""

import os
import json
import base64
import secrets
import hashlib
import struct
import time
from io import BytesIO
from PIL import Image
import numpy as np
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
import logging

from .exceptions import (
    EncryptionError,
    DecryptionError,
    InvalidPasswordError,
    InvalidKeyError,
    UnsupportedFormatError
)
from .encryption_v4 import CryptoPixV4

logger = logging.getLogger(__name__)

# CryptoPix V4 Configuration - Ultra-Fast Performance
PBKDF2_ITERATIONS = 1000  # Optimized for ultra-fast performance while maintaining security
KEY_LENGTH = 32  # 256 bits
SALT_LENGTH = 16  # 128 bits
CHUNK_SIZE = 24  # 24-bit chunks for RGB colors
VERSION = "4.0"


class CryptoPix:
    """Main class for CryptoPix V4 ultra-fast encryption and decryption operations with backward compatibility"""
    
    def __init__(self):
        """Initialize CryptoPix V4 with ultra-fast engine"""
        self.v4_engine = CryptoPixV4()
        # Keep color table for backward compatibility
        self.color_table = self._generate_default_color_table()
    
    def _generate_default_color_table(self):
        """Generate a minimal color mapping table - colors generated on demand"""
        return {}
    
    def _derive_key(self, password: str, salt: bytes) -> bytes:
        """
        Derive a 256-bit key from password using PBKDF2-HMAC-SHA256
        
        Args:
            password: User-provided password
            salt: 128-bit random salt
            
        Returns:
            256-bit derived key
            
        Raises:
            EncryptionError: If key derivation fails
        """
        try:
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=KEY_LENGTH,
                salt=salt,
                iterations=PBKDF2_ITERATIONS,
                backend=default_backend()
            )
            return kdf.derive(password.encode('utf-8'))
        except Exception as e:
            raise EncryptionError(f"Failed to derive key: {str(e)}")
    
    def _generate_color_for_chunk(self, chunk: str, key: bytes) -> tuple:
        """
        Generate RGB color for a binary chunk using key-based deterministic method
        
        Args:
            chunk: 24-bit binary string
            key: Derived key for color generation
            
        Returns:
            RGB tuple (r, g, b)
        """
        # Use direct binary to RGB conversion for simplicity and efficiency
        r = int(chunk[0:8], 2)
        g = int(chunk[8:16], 2)
        b = int(chunk[16:24], 2)
        
        # Apply key-based transformation for security
        key_bytes = key[:3]
        r = (r + key_bytes[0]) % 256
        g = (g + key_bytes[1]) % 256
        b = (b + key_bytes[2]) % 256
        
        return (r, g, b)
    
    def _encrypt_metadata(self, metadata: dict, key: bytes) -> str:
        """
        Encrypt metadata using AES-256-GCM and encode as base64
        
        Args:
            metadata: Metadata dictionary to encrypt
            key: Encryption key
            
        Returns:
            Base64-encoded encrypted metadata
            
        Raises:
            EncryptionError: If metadata encryption fails
        """
        try:
            # Pack metadata into binary format instead of JSON for better performance
            # Format: version(2 bytes) + chunk_count(4 bytes) + padding(1 byte) + shuffle_seed(16 bytes)
            version_bytes = VERSION.encode('utf-8')[:2].ljust(2, b'\x00')
            chunk_count = metadata['chunk_count']
            padding = metadata['padding']
            shuffle_seed = base64.b64decode(metadata['shuffle_seed'])
            
            # Pack into binary struct: 2s I B 16s (total 23 bytes)
            metadata_binary = struct.pack('2sIB16s', version_bytes, chunk_count, padding, shuffle_seed)
            
            # Generate random IV
            iv = os.urandom(12)  # 96-bit IV for GCM
            
            # Encrypt using AES-256-GCM
            cipher = Cipher(algorithms.AES(key), modes.GCM(iv), backend=default_backend())
            encryptor = cipher.encryptor()
            ciphertext = encryptor.update(metadata_binary) + encryptor.finalize()
            
            # Combine IV + ciphertext + tag
            encrypted_data = iv + ciphertext + encryptor.tag
            
            # Encode as base64
            return base64.b64encode(encrypted_data).decode('utf-8')
        except Exception as e:
            raise EncryptionError(f"Failed to encrypt metadata: {str(e)}")
    
    def _decrypt_metadata(self, encrypted_metadata: str, key: bytes) -> dict:
        """
        Decrypt metadata from base64-encoded encrypted string
        
        Args:
            encrypted_metadata: Base64-encoded encrypted metadata
            key: Decryption key
            
        Returns:
            Decrypted metadata dictionary
            
        Raises:
            DecryptionError: If metadata decryption fails
            InvalidPasswordError: If password is incorrect
        """
        try:
            # Decode from base64
            encrypted_data = base64.b64decode(encrypted_metadata)
            
            # Extract components
            iv = encrypted_data[:12]  # 96-bit IV
            tag = encrypted_data[-16:]  # 128-bit tag
            ciphertext = encrypted_data[12:-16]
            
            # Decrypt using AES-256-GCM
            cipher = Cipher(algorithms.AES(key), modes.GCM(iv, tag), backend=default_backend())
            decryptor = cipher.decryptor()
            decrypted_data = decryptor.update(ciphertext) + decryptor.finalize()
            
            # Unpack binary struct: 2s I B 16s
            version_bytes, chunk_count, padding, shuffle_seed = struct.unpack('2sIB16s', decrypted_data)
            
            # Convert back to dictionary format
            return {
                'version': version_bytes.rstrip(b'\x00').decode('utf-8'),
                'chunk_count': chunk_count,
                'padding': padding,
                'shuffle_seed': base64.b64encode(shuffle_seed).decode('utf-8')
            }
            
        except Exception as e:
            error_message = str(e).lower()
            if "authentication" in error_message or "tag" in error_message or "invalid" in error_message:
                raise InvalidPasswordError("Incorrect password or corrupted metadata")
            raise DecryptionError(f"Failed to decrypt metadata: {str(e)}")
    
    def encrypt(self, text: str, password: str, width=None) -> tuple:
        """
        Encrypt text into a WebP image using ultra-fast CryptoPix V4 algorithm
        
        Args:
            text: Plain-text data to encrypt
            password: User-provided password
            width: Optional image width (auto-calculated if None)
            
        Returns:
            Tuple of (BytesIO image, smart_key string)
            
        Raises:
            EncryptionError: If encryption process fails
            ValueError: If input parameters are invalid
        """
        # Use ultra-fast V4 engine for all new encryptions
        return self.v4_engine.encrypt(text, password, width)
    
    def decrypt(self, img: Image.Image, smart_key: str, password: str) -> dict:
        """
        Decrypt an encrypted WebP image back to text (supports V2 and V4 keys)
        
        Args:
            img: PIL Image object
            smart_key: Smart key containing encrypted metadata
            password: Same password used for encryption
            
        Returns:
            Dictionary with decrypted content and type
            
        Raises:
            DecryptionError: If decryption process fails
            InvalidKeyError: If smart key is invalid
            InvalidPasswordError: If password is incorrect
        """
        if not smart_key:
            raise InvalidKeyError("Smart key cannot be empty")
        if not password:
            raise ValueError("Password cannot be empty")
        
        # Detect key version and route to appropriate engine
        if smart_key.startswith("cryptopix_v4:"):
            # Use ultra-fast V4 engine
            return self.v4_engine.decrypt(img, smart_key, password)
        elif smart_key.startswith("cryptopix_v2:"):
            # Use legacy V2 decryption for backward compatibility
            return self._decrypt_v2_legacy(img, smart_key, password)
        else:
            raise InvalidKeyError("Unsupported smart key format")
    
    def encrypt_fast(self, text: str, password: str) -> tuple:
        """
        Ultra-fast raw encryption without image conversion (Speed Mode)
        
        Args:
            text: Plain-text data to encrypt
            password: User-provided password
            
        Returns:
            Tuple of (encrypted_bytes, key_data_string)
            
        Raises:
            EncryptionError: If encryption process fails
            ValueError: If input parameters are invalid
        """
        if not text:
            raise ValueError("Text to encrypt cannot be empty")
        if not password:
            raise ValueError("Password cannot be empty")
            
        try:
            # Generate salt and derive key
            salt = os.urandom(SALT_LENGTH)
            derived_key = self._derive_key(password, salt)
            
            # Convert text to bytes
            text_bytes = text.encode('utf-8')
            
            # Generate random IV for AES-GCM
            iv = os.urandom(12)  # 96-bit IV for GCM
            
            # Encrypt using AES-256-GCM
            cipher = Cipher(algorithms.AES(derived_key), modes.GCM(iv), backend=default_backend())
            encryptor = cipher.encryptor()
            ciphertext = encryptor.update(text_bytes) + encryptor.finalize()
            
            # Combine IV + ciphertext + tag
            encrypted_data = iv + ciphertext + encryptor.tag
            
            # Create key data string with salt and metadata
            key_data = f"cryptopix_fast_v4:{base64.b64encode(salt).decode('utf-8')}"
            
            return encrypted_data, key_data
            
        except Exception as e:
            raise EncryptionError(f"Fast encryption failed: {str(e)}")
    
    def decrypt_fast(self, encrypted_data: bytes, key_data: str, password: str) -> str:
        """
        Ultra-fast raw decryption from encrypted bytes (Speed Mode)
        
        Args:
            encrypted_data: Encrypted byte data
            key_data: Key data string from encryption
            password: Same password used for encryption
            
        Returns:
            Decrypted text string
            
        Raises:
            DecryptionError: If decryption process fails
            InvalidKeyError: If key data is invalid
            InvalidPasswordError: If password is incorrect
        """
        if not encrypted_data:
            raise ValueError("Encrypted data cannot be empty")
        if not key_data:
            raise InvalidKeyError("Key data cannot be empty")
        if not password:
            raise ValueError("Password cannot be empty")
            
        try:
            # Parse key data
            if not key_data.startswith("cryptopix_fast_v4:"):
                raise InvalidKeyError("Invalid key data format for fast mode")
            
            salt_b64 = key_data.split(":", 1)[1]
            salt = base64.b64decode(salt_b64)
            
            # Derive key
            derived_key = self._derive_key(password, salt)
            
            # Extract components from encrypted data
            iv = encrypted_data[:12]  # 96-bit IV
            tag = encrypted_data[-16:]  # 128-bit tag
            ciphertext = encrypted_data[12:-16]
            
            # Decrypt using AES-256-GCM
            cipher = Cipher(algorithms.AES(derived_key), modes.GCM(iv, tag), backend=default_backend())
            decryptor = cipher.decryptor()
            decrypted_bytes = decryptor.update(ciphertext) + decryptor.finalize()
            
            # Convert back to text
            return decrypted_bytes.decode('utf-8')
            
        except Exception as e:
            error_message = str(e).lower()
            if "authentication" in error_message or "tag" in error_message or "invalid" in error_message:
                raise InvalidPasswordError("Incorrect password or corrupted data")
            raise DecryptionError(f"Fast decryption failed: {str(e)}")

    def _decrypt_v2_legacy(self, img: Image.Image, smart_key: str, password: str) -> dict:
        """Legacy V2 decryption for backward compatibility"""
        try:
            # Parse the format: cryptopix_v2:salt_b64:encrypted_metadata
            key_parts = smart_key.split(':', 2)
            if len(key_parts) != 3:
                raise InvalidKeyError("Invalid smart key format - missing components")
            
            _, salt_b64, encrypted_metadata = key_parts
            
            # Step 2: Extract salt and derive key
            try:
                salt = base64.b64decode(salt_b64)
                derived_key = self._derive_key(password, salt)
                
                # Decrypt metadata with proper key
                metadata = self._decrypt_metadata(encrypted_metadata, derived_key)
                
            except (InvalidPasswordError, DecryptionError):
                raise
            except Exception as e:
                raise DecryptionError(f"Failed to process key: {str(e)}")
            
            # Step 3: Extract metadata
            chunk_count = metadata['chunk_count']
            padding = metadata['padding']
            
            # Step 4: Ultra-fast vectorized pixel extraction using NumPy
            img_array = np.asarray(img)
            pixel_data = img_array.reshape(-1, 3)
            
            # Extract required pixels and reverse key transformation
            pixels_subset = pixel_data[:chunk_count]
            
            # Vectorized key reversal
            key_bytes = np.frombuffer(derived_key[:3], dtype=np.uint8)
            decoded_pixels = (pixels_subset.astype(np.int16) - key_bytes) % 256
            decoded_pixels = decoded_pixels.astype(np.uint8)
            
            # Direct conversion back to bytes - no binary string processing
            text_bytes_flat = decoded_pixels.flatten()
            
            # Remove padding if present
            if padding > 0:
                text_bytes = text_bytes_flat[:-padding]
            else:
                text_bytes = text_bytes_flat
            
            # Step 5: Decode to text
            try:
                decrypted_text = bytes(text_bytes).decode('utf-8')
                return {
                    'content': decrypted_text,
                    'type': 'text',
                    'success': True
                }
            except UnicodeDecodeError:
                return {
                    'content': base64.b64encode(bytes(text_bytes)).decode('utf-8'),
                    'type': 'binary',
                    'success': True
                }
                
        except (DecryptionError, InvalidKeyError, InvalidPasswordError, ValueError):
            raise
        except Exception as e:
            raise DecryptionError(f"Decryption failed: {str(e)}")


# Convenience functions for backward compatibility
def encrypt_text_to_image_v2(text: str, password: str, width=None) -> tuple:
    """
    Encrypt text to image using CryptoPix V2 algorithm
    
    Args:
        text: Text to encrypt
        password: Password for encryption
        width: Optional image width
        
    Returns:
        Tuple of (BytesIO image, smart_key)
    """
    cp = CryptoPix()
    return cp.encrypt(text, password, width)


def decrypt_image_to_text_v2(img: Image.Image, smart_key: str, password: str) -> dict:
    """
    Decrypt image to text using CryptoPix V2 algorithm
    
    Args:
        img: PIL Image object
        smart_key: Smart key with encrypted metadata
        password: Password for decryption
        
    Returns:
        Dictionary with decrypted content
    """
    cp = CryptoPix()
    return cp.decrypt(img, smart_key, password)